using CliWrap;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.Logging;
using ShellProgressBar;
using System.Collections.ObjectModel;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using View_by_Distance.Metadata.Models;
using View_by_Distance.Metadata.Models.Stateless.Methods;
using View_by_Distance.Rename.Models;
using View_by_Distance.Shared.Models;
using View_by_Distance.Shared.Models.Properties;
using View_by_Distance.Shared.Models.Stateless.Methods;

namespace View_by_Distance.Rename;

public class Rename : IRename
{

    private record ToDo(string? Directory, FileHolder FileHolder, string File, bool JsonFile);
    private record Record(DateTime DateTime, ExifDirectory ExifDirectory, string File, string JsonFile);

    private ProgressBar? _ProgressBar;

    public Rename(List<string> args, ILogger<Program>? logger, IConfigurationRoot configurationRoot, AppSettings appSettings, bool isSilent, IConsole console)
    {
        if (isSilent)
        { }
        if (args is null)
            throw new NullReferenceException(nameof(args));
        if (console is null)
            throw new NullReferenceException(nameof(console));
        IRename rename = this;
        long ticks = DateTime.Now.Ticks;
        ResultConfiguration resultConfiguration = Metadata.Models.Binder.ResultConfiguration.Get(configurationRoot, appSettings.RequireRootDirectoryExists);
        MetadataConfiguration metadataConfiguration = Metadata.Models.Binder.MetadataConfiguration.Get(configurationRoot, resultConfiguration);
        RenameConfiguration renameConfiguration = Models.Binder.RenameConfiguration.Get(configurationRoot, metadataConfiguration);
        RenameWork(logger, appSettings, rename, ticks, renameConfiguration);
    }

    void IRename.Tick() =>
        _ProgressBar?.Tick();

    (ReadOnlyCollection<string>, FilePath?) IRename.ConvertAndGetFfmpegFiles(IRenameConfiguration renameConfiguration, FilePath filePath)
    {
        List<string> results = [];
        FilePath? result;
        bool isIgnoreExtension;
        bool isValidImageFormatExtension = renameConfiguration.ValidImageFormatExtensions.Contains(filePath.ExtensionLowered);
        isIgnoreExtension = isValidImageFormatExtension && renameConfiguration.IgnoreExtensions.Contains(filePath.ExtensionLowered);
        if (!isIgnoreExtension && isValidImageFormatExtension)
            result = null;
        else
        {
            CommandTask<CommandResult> commandTask = Cli.Wrap("ffmpeg.exe")
                .WithArguments(new[] { "-i", filePath.FullName, "-vf", "select=eq(n\\,0)", "-q:v", "1", $"{filePath.Name}-%4d.jpg" })
                .WithWorkingDirectory(filePath.DirectoryName)
                .ExecuteAsync();
            commandTask.Task.Wait();
            results.AddRange(Directory.GetFiles(filePath.DirectoryName, $"{filePath.Name}-*.jpg", SearchOption.TopDirectoryOnly));
            if (results.Count == 0)
                throw new Exception();
            result = IId.GetFilePath(renameConfiguration.MetadataConfiguration, results[0], index: null);
            if (!result.Name.EndsWith("-0001.jpg"))
                throw new Exception();
            isValidImageFormatExtension = renameConfiguration.ValidImageFormatExtensions.Contains(result.ExtensionLowered);
            isIgnoreExtension = isValidImageFormatExtension && renameConfiguration.IgnoreExtensions.Contains(result.ExtensionLowered);
            if (isIgnoreExtension || !isValidImageFormatExtension)
                throw new Exception();
            if (result.DirectoryName is null)
                throw new NullReferenceException(nameof(result.DirectoryName));
        }
        return new(new(results), result);
    }

#pragma warning disable CA1416

    DeterministicHashCode IRename.GetDeterministicHashCode(FilePath filePath)
    {
        DeterministicHashCode result;
        int? id;
        int? width;
        int? height;
        try
        {
            using Image image = Image.FromFile(filePath.FullName);
            width = image.Width;
            height = image.Height;
            using Bitmap bitmap = new(image);
            Rectangle rectangle = new(0, 0, image.Width, image.Height);
            BitmapData bitmapData = bitmap.LockBits(rectangle, ImageLockMode.ReadOnly, bitmap.PixelFormat);
            IntPtr intPtr = bitmapData.Scan0;
            int length = bitmapData.Stride * bitmap.Height;
            byte[] bytes = new byte[length];
            Marshal.Copy(intPtr, bytes, 0, length);
            bitmap.UnlockBits(bitmapData);
            id = IId.GetDeterministicHashCode(bytes);
        }
        catch (Exception)
        {
            id = null;
            width = null;
            height = null;
        }
        result = new(height, id, width);
        return result;
    }

#pragma warning restore CA1416

    private static void GetExifDirectoryCollection(IRename rename, RenameConfiguration renameConfiguration, List<(string, FileInfo, ExifDirectory)> exifDirectories, IEnumerable<string> files, A_Metadata metadata)
    {
        int index = -1;
        FileInfo fileInfo;
        FilePath filePath;
        FilePath? ffmpegFilePath;
        ExifDirectory exifDirectory;
        ReadOnlyCollection<string> ffmpegFiles;
        DeterministicHashCode deterministicHashCode;
        foreach (string file in files)
        {
            index += 1;
            rename.Tick();
            filePath = IId.GetFilePath(renameConfiguration.MetadataConfiguration, file, index);
            if (renameConfiguration.SkipIdFiles && filePath.Id is not null && (filePath.IsIntelligentIdFormat || filePath.SortOrder is not null))
                continue;
            (ffmpegFiles, ffmpegFilePath) = rename.ConvertAndGetFfmpegFiles(renameConfiguration, filePath);
            if (ffmpegFilePath is not null)
                filePath = ffmpegFilePath;
            if (filePath.Id is not null)
                deterministicHashCode = new(null, filePath.Id, null);
            else
                deterministicHashCode = rename.GetDeterministicHashCode(filePath);
            (fileInfo, exifDirectory) = metadata.GetMetadataCollection(renameConfiguration.MetadataConfiguration, filePath, deterministicHashCode);
            exifDirectories.Add(new(file, fileInfo, exifDirectory));
            foreach (string ffmpegFile in ffmpegFiles)
                File.Delete(ffmpegFile);
        }
    }

    private static ReadOnlyCollection<Record> GetExifDirectoryCollection(List<(string, FileInfo, ExifDirectory)> exifDirectories)
    {
        List<Record> results = [];
        DateTime? dateTime;
        foreach ((string file, FileInfo fileInfo, ExifDirectory exifDirectory) in exifDirectories)
        {
            dateTime = IDate.GetDateTimeOriginal(exifDirectory);
            dateTime ??= IDate.GetMinimum(exifDirectory);
            results.Add(new(dateTime.Value, exifDirectory, file, fileInfo.FullName));
        }
        return new(results);
    }

    private ReadOnlyCollection<Record> GetExifDirectoryCollection(IRename rename, AppSettings appSettings, RenameConfiguration renameConfiguration, DirectoryInfo directoryInfo)
    {
        ReadOnlyCollection<Record> results;
        List<(string, FileInfo, ExifDirectory)> exifDirectories = [];
        int appSettingsMaxDegreeOfParallelism = appSettings.MaxDegreeOfParallelism;
        IEnumerable<string> files = Directory.EnumerateFiles(directoryInfo.FullName, "*", SearchOption.AllDirectories);
        A_Metadata metadata = new(renameConfiguration.MetadataConfiguration);
        _ProgressBar = new(123000, "EnumerateFiles load", new ProgressBarOptions() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true });
        if (appSettingsMaxDegreeOfParallelism == 1)
            GetExifDirectoryCollection(rename, renameConfiguration, exifDirectories, files, metadata);
        else
        {
            ParallelOptions parallelOptions = new() { MaxDegreeOfParallelism = appSettingsMaxDegreeOfParallelism };
            files.AsParallel().ForAll(A_Metadata.SetExifDirectoryCollection(rename, renameConfiguration, metadata, exifDirectories));
            if (_ProgressBar.CurrentTick != exifDirectories.Count)
                throw new NotSupportedException();
        }
        _ProgressBar.Dispose();
        results = GetExifDirectoryCollection(exifDirectories);
        return results;
    }

    private static void VerifyIntMinValueLength(MetadataConfiguration metadataConfiguration, ReadOnlyCollection<Record> exifDirectories)
    {
        foreach ((DateTime _, ExifDirectory exifDirectory, string _, string _) in exifDirectories)
        {
            if (exifDirectory.Id is null)
                continue;
            if (metadataConfiguration.IntMinValueLength < exifDirectory.Id.Value.ToString().Length)
                throw new NotSupportedException();
        }
    }

    private static string GetCheckDirectory(RenameConfiguration renameConfiguration, Record record, FileHolder fileHolder)
    {
        string? checkDirectory;
        if (fileHolder.DirectoryName is null)
            throw new NullReferenceException(nameof(fileHolder.DirectoryName));
        string directoryName;
        string year = record.DateTime.Year.ToString();
        string checkDirectoryName = Path.GetFileName(fileHolder.DirectoryName);
        if (!checkDirectoryName.Contains(year))
            throw new NotImplementedException();
        else
        {
            string maker = IMetadata.GetMaker(record.ExifDirectory.ExifDirectoryBase);
            (int seasonValue, string seasonName) = IDate.GetSeason(record.DateTime.DayOfYear);
            string splat = fileHolder.DirectoryName[^3..][1] == '!' ? fileHolder.DirectoryName[^3..] : string.Empty;
            directoryName = $"{year}.{seasonValue} {seasonName} {maker.Split(' ')[0]}{splat}";
        }
        string rootDirectory = renameConfiguration.MoveFilesToRoot ? renameConfiguration.MetadataConfiguration.ResultConfiguration.RootDirectory : fileHolder.DirectoryName;
        checkDirectory = Path.Combine(rootDirectory, "_ Destination _", directoryName);
        return checkDirectory;
    }

    private static ReadOnlyCollection<ToDo> GetToDoCollection(ILogger<Program>? logger, RenameConfiguration renameConfiguration, long ticks, ReadOnlyCollection<Record> exifDirectories)
    {
        List<ToDo> results = [];
        Record record;
        string jsonFile;
        string paddedId;
        string checkFile;
        string directoryName;
        FileHolder fileHolder;
        string? checkDirectory;
        const string jpg = ".jpg";
        string checkFileExtension;
        List<string> distinct = [];
        const string jpeg = ".jpeg";
        string jsonFileSubDirectory;
        VerifyIntMinValueLength(renameConfiguration.MetadataConfiguration, exifDirectories);
        ResultConfiguration resultConfiguration = renameConfiguration.MetadataConfiguration.ResultConfiguration;
        ReadOnlyCollection<Record> records = new((from l in exifDirectories orderby l.DateTime select l).ToArray());
        for (int i = 0; i < records.Count; i++)
        {
            record = records[i];
            if (record.ExifDirectory.Id is null)
                continue;
            fileHolder = new(record.File);
            if (fileHolder.DirectoryName is null)
                continue;
            checkDirectory = GetCheckDirectory(renameConfiguration, record, fileHolder);
            checkFileExtension = fileHolder.ExtensionLowered == jpeg ? jpg : fileHolder.ExtensionLowered;
            paddedId = IId.GetPaddedId(renameConfiguration.MetadataConfiguration, i, record.ExifDirectory.Id.Value);
            jsonFileSubDirectory = Path.GetDirectoryName(Path.GetDirectoryName(record.JsonFile)) ?? throw new Exception();
            checkFile = Path.Combine(checkDirectory, $"{paddedId}{checkFileExtension}");
            if (checkFile == fileHolder.FullName)
                continue;
            if (File.Exists(checkFile))
            {
                checkFile = string.Concat(checkFile, ".del");
                if (File.Exists(checkFile))
                    continue;
            }
            (directoryName, _) = IPath.GetDirectoryNameAndIndex(resultConfiguration, record.ExifDirectory.Id.Value);
            jsonFile = Path.Combine(jsonFileSubDirectory, directoryName, $"{record.ExifDirectory.Id.Value}{checkFileExtension}.json");
            if (record.JsonFile != jsonFile)
                results.Add(new(null, new(record.JsonFile), jsonFile, JsonFile: true));
            if (distinct.Contains(checkFile))
                continue;
            distinct.Add(checkFile);
            results.Add(new(checkDirectory, fileHolder, checkFile, JsonFile: false));
        }
        return new(results);
    }

    private static void VerifyDirectories(ReadOnlyCollection<ToDo> toDoCollection)
    {
        List<string> distinct = [];
        foreach (ToDo toDo in toDoCollection)
        {
            if (toDo.Directory is null || distinct.Contains(toDo.Directory))
                continue;
            if (!Directory.Exists(toDo.Directory))
                _ = Directory.CreateDirectory(toDo.Directory);
            distinct.Add(toDo.Directory);
        }
    }

    private ReadOnlyCollection<string> RenameFilesInDirectories(ReadOnlyCollection<ToDo> toDoCollection)
    {
        List<string> results = [];
        VerifyDirectories(toDoCollection);
        _ProgressBar = new(toDoCollection.Count, "Move Files", new ProgressBarOptions() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true });
        foreach (ToDo toDo in toDoCollection)
        {
            _ProgressBar.Tick();
            if (toDo.JsonFile)
            {
                if (File.Exists(toDo.File))
                    File.Delete(toDo.File);
                File.Move(toDo.FileHolder.FullName, toDo.File);
            }
            else if (toDo.Directory is null)
                throw new NotSupportedException();
            else
            {
                if (File.Exists(toDo.File))
                    File.Delete(toDo.File);
                try
                { File.Move(toDo.FileHolder.FullName, toDo.File); }
                catch (Exception)
                { continue; }
                results.Add($"{toDo.FileHolder.FullName}\t{toDo.File}");
            }
        }
        _ProgressBar.Dispose();
        return new(results);
    }

    private void RenameWork(ILogger<Program>? logger, AppSettings appSettings, IRename rename, long ticks, RenameConfiguration renameConfiguration)
    {
        DirectoryInfo directoryInfo = new(Path.GetFullPath(renameConfiguration.MetadataConfiguration.ResultConfiguration.RootDirectory));
        logger?.LogInformation("{RootDirectory}", directoryInfo.FullName);
        ReadOnlyCollection<Record> exifDirectories = GetExifDirectoryCollection(rename, appSettings, renameConfiguration, directoryInfo);
        ReadOnlyCollection<ToDo> toDoCollection = GetToDoCollection(logger, renameConfiguration, ticks, exifDirectories);
        ReadOnlyCollection<string> lines = RenameFilesInDirectories(toDoCollection);
        if (lines.Count != 0)
        {
            File.WriteAllLines($"D:/Tmp/Phares/{DateTime.Now.Ticks}.tsv", lines);
            _ = IPath.DeleteEmptyDirectories(directoryInfo.FullName);
        }
    }

}